/*
    Drools5 Integration Helper
    Copyright (C) 2009  Mathieu Boretti mathieu.boretti@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */
package org.boretti.drools.integration.drools5.implementation;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.lang.reflect.Array;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.log4j.Logger;
import org.boretti.drools.integration.drools5.DroolsInterface;
import org.boretti.drools.integration.drools5.exceptions.IODroolsError;
import org.boretti.drools.integration.drools5.exceptions.MalformedURLDroolsError;
import org.drools.RuleBase;
import org.drools.StatefulSession;
import org.drools.StatelessSession;
import org.drools.io.Resource;
import org.drools.io.ResourceFactory;

/**
 * 
 * @author mbo
 *
 */
abstract class DroolsInterfaceImplementation<T> implements DroolsInterface{
	
	private static final Logger logger = Logger.getLogger(DroolsInterfaceImplementation.class);
	
	private Class<T> clazz; 
		
	private RuleBase ruleBase;
	
	private StatefulSession session;
	
	private static final String FIELD_NAME_LOGGER = "droolsLogger";
	
	private Map<String,DroolsVariableImplementation> fields = new HashMap<String,DroolsVariableImplementation>();

	@Override
	public RuleBase getRuleBase() {
		return ruleBase;
	}
	
	protected DroolsInterfaceImplementation(Class<T> clazz,RuleBase ruleBase) {
		this.clazz=clazz;
		this.ruleBase=ruleBase;
		fields.put(FIELD_NAME_LOGGER,new DroolsVariableImplementation(true,Logger.getLogger(clazz)));
	}
	
	protected static Resource getAsDroolsResource(Class<?> clazz,String resourceName) {
		return ResourceFactory.newReaderResource(getReaderFromResource(clazz,resourceName));
	}
	
	protected static Reader getReaderFromResource(Class<?> clazz,String resourceName) {
		return new InputStreamReader(getInputStreamFromResource(clazz,resourceName));
	}
	
	protected static InputStream getInputStreamFromResource(Class<?> clazz,String resourceName) {
		resourceName = resourceName.trim();
		InputStream is=null;
		if (resourceName.startsWith("http:/")||resourceName.startsWith("file:/")) {
			try {
				URL url = new URL(resourceName);
				is = url.openStream();
			} catch (MalformedURLException e) {
				String msg = "URL error ("+resourceName+"):"+e.getMessage();
				logger.error(msg);
				throw new MalformedURLDroolsError(msg);
			} catch (IOException e) {
				String msg = "URL error ("+resourceName+"):"+e.getMessage();
				logger.error(msg);
				throw new IODroolsError(msg);
			}
		} else {
			is = clazz.getResourceAsStream(resourceName);
		}
		if (is==null) {
			String msg = "IO Exception, resource not found "+resourceName;
			logger.error(msg);
			throw new IODroolsError(msg);
		}
		return is;
	}
	

	public Class<T> getClazz() {
		return clazz;
	}

	@Override
	public StatefulSession getCurrentSession() {
		return session;
	}
	
	void setCurrentSession(StatefulSession session) {
		this.session=session;
	}
	
	Iterator<?> runStatelessSession(List<Object> fact) {
		StatelessSession dss = getRuleBase().newStatelessSession();
		for(Entry<String,DroolsVariableImplementation> es:fields.entrySet()) {
			String field = es.getKey();
			DroolsVariableImplementation data = es.getValue();
			if (data.isGlobal()) dss.setGlobal(field, data.getData());
			else fact.add(data.getData());
		}
		new DroolsWorkingMemoryLogger(dss);
		return dss.executeWithResults(fact).iterateObjects();
	}
	
	Iterator<?> runStatefulSession(List<Object> fact,boolean requiredExisting) {
		if (requiredExisting && session==null)  {
			String msg = "No session found when required";
			logger.error(msg);
			throw new IllegalArgumentException(msg);
		}
		if (session==null) {
			session=ruleBase.newStatefulSession();
			for(Entry<String,DroolsVariableImplementation> es:fields.entrySet()) {
				String field = es.getKey();
				DroolsVariableImplementation data = es.getValue();
				if (data.isGlobal()) session.setGlobal(field, data.getData());
				else session.insert(data.getData());
			}
			new DroolsWorkingMemoryLogger(session);
		}
		for(Object o:fact) session.insert(o);
		session.fireAllRules();
		return session.iterateObjects();
	}
	
	Object getResult(Class<?> returnType,Iterator<?> result) {
		return getInternalResult(returnType,result);
	}
	
	Object getInternalResult(Class<?> returnType,Iterator<?> result) {
		if (returnType.equals(Void.TYPE)) return null;
		if (logger.isDebugEnabled()) logger.debug("Type is "+returnType);
		if (result!=null) {
			if (returnType.isArray()) {
				Class<?> real = returnType.getComponentType();
				if (logger.isDebugEnabled()) logger.debug("Type is array of "+real);
				List<Object> lst = new ArrayList<Object>();
				Iterator<?> i = result;
				while(i.hasNext()) {
					Object o = i.next();
					if (logger.isDebugEnabled()) logger.debug("In evaluation object is "+o);
					if (isTypeCompatible(real,o.getClass())) lst.add(o);
				}
				return lst.toArray((Object[])Array.newInstance(real,lst.size()));
			} else {
				Iterator<?> i = result;
				while(i.hasNext()) {
					Object o = i.next();
					if (logger.isDebugEnabled()) logger.debug("In evaluation object is "+o);
					if (isTypeCompatible(returnType,o.getClass())) return o;
				}
			}
		}
		return null;
	}
	
	/**
	 * 
	 * @param type
	 * @param src
	 * @return
	 */
	
	private boolean isTypeCompatible(Class<?> type,Class<?> src) {
		if (logger.isDebugEnabled()) logger.debug("type is "+type+" ; src is "+src);
		if (type.equals(src)) return true;
		if (type.isAssignableFrom(src)) return true;
		if (type.isPrimitive()) {
			if (type.equals(Boolean.TYPE) && src.equals(Boolean.class)) return true;
			if (src.equals(Boolean.TYPE) && type.equals(Boolean.class)) return true;
			if (type.equals(Integer.TYPE) && src.equals(Integer.class)) return true;
			if (src.equals(Integer.TYPE) && type.equals(Integer.class)) return true;
			if (type.equals(Long.TYPE) && src.equals(Long.class)) return true;
			if (src.equals(Long.TYPE) && type.equals(Long.class)) return true;
			if (type.equals(Short.TYPE) && src.equals(Short.class)) return true;
			if (src.equals(Short.TYPE) && type.equals(Short.class)) return true;
			if (type.equals(Byte.TYPE) && src.equals(Byte.class)) return true;
			if (src.equals(Byte.TYPE) && type.equals(Byte.class)) return true;
			if (type.equals(Float.TYPE) && src.equals(Float.class)) return true;
			if (src.equals(Float.TYPE) && type.equals(Float.class)) return true;
			if (type.equals(Double.TYPE) && src.equals(Double.class)) return true;
			if (src.equals(Double.TYPE) && type.equals(Double.class)) return true;
			if (type.equals(Character.TYPE) && src.equals(Character.class)) return true;
			if (src.equals(Character.TYPE) && type.equals(Character.class)) return true;
		}
		return false;
	}

	/**
	 * @return the fields
	 */
	Map<String, DroolsVariableImplementation> getFields() {
		return fields;
	}

	/* (non-Javadoc)
	 * @see org.boretti.drools.integration.drools5.DroolsInterface#getFieldByName(java.lang.String, java.lang.Class)
	 */
	@SuppressWarnings("unchecked")
	@Override
	public <T1> T1 getFieldByName(String name, Class<T1> clazz) {
		DroolsVariableImplementation dvi = fields.get(name);
		if (dvi==null) return null;
		if (isTypeCompatible(clazz,dvi.getData().getClass())) return (T1)dvi.getData();
		String msg = "Expected "+clazz+"; found "+dvi.getClass().getClass();
		logger.error(msg);
		throw new IllegalArgumentException(msg);
	}

}