/*
    Drools5 Integration Helper
    Copyright (C) 2009  Mathieu Boretti mathieu.boretti@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */
package org.boretti.drools.integration.drools5.implementation;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;

import org.apache.log4j.Logger;
import org.boretti.drools.integration.drools5.DroolsInterface;
import org.boretti.drools.integration.drools5.DroolsSessionType;
import org.boretti.drools.integration.drools5.annotations.DroolsMethod;
import org.boretti.drools.integration.drools5.annotations.DroolsVariable;
import org.boretti.drools.integration.drools5.exceptions.IncompatibleAnnotationDroolsError;

/**
 * @author mbo
 * 
 */
class DroolsProxy<T> implements InvocationHandler {

	private final Logger logger = Logger.getLogger(this.getClass());

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.reflect.InvocationHandler#invoke(java.lang.Object,
	 *      java.lang.reflect.Method, java.lang.Object[])
	 */
	@Override
	public Object invoke(final Object proxy, final Method method,
			final Object[] args) throws Throwable {
		if (method.getDeclaringClass().equals(clazz)) {
			DroolsMethod annotation = null;
			DroolsVariable variable = null;
			try {
				annotation = method.getAnnotation(DroolsMethod.class);
			} catch (NullPointerException ex) {
			}
			try {
				variable = method.getAnnotation(DroolsVariable.class);
			} catch (NullPointerException ex) {
			}
			if (annotation != null && variable != null) {
				String msg = "Method is annotated with both @DroolsMethod and @DroolsVariable.";
				logger.error(msg);
				throw new IncompatibleAnnotationDroolsError(msg);
			}
			if (variable != null) {
				String methodName = method.getName();
				if (!methodName.startsWith("set")) {
					String msg = "Method is annotated with @DroolsVariable and name is not set<something>.";
					logger.error(msg);
					throw new IncompatibleAnnotationDroolsError(msg);
				}
				if (method.getParameterTypes().length != 1) {
					String msg = "Method is annotated with @DroolsVariable and name is set<something>, but number of parameter is not 1.";
					logger.error(msg);
					throw new IncompatibleAnnotationDroolsError(msg);
				}
				if (!method.getReturnType().equals(Void.TYPE)) {
					String msg = "Method is annotated with @DroolsVariable and name is set<something> and number of parameter is 1, but return type is not void.";
					logger.error(msg);
					throw new IncompatibleAnnotationDroolsError(msg);
				}
				String fieldName = methodName.substring(3, 4).toLowerCase()
						+ methodName.substring(4);
				Object data = args[0];
				Map<String, DroolsVariableImplementation> fields = inter
						.getFields();
				if (data == null) {
					if (logger.isDebugEnabled())
						logger
								.debug("Received parameter is null. Retracting field for next call");
					if (fields.containsKey(fieldName))
						fields.remove(fieldName);
				} else {
					fields.put(fieldName, new DroolsVariableImplementation(
							variable.global(), data));
				}
				return null;
			} else {
				Class<?> c = method.getReturnType();
				DroolsSessionType session = DroolsSessionType.NONE;
				if (annotation != null && annotation.session() != null)
					session = annotation.session();
				DroolsProductFactResult facts = new DroolsProductFactResult(
						method, args);
				List<Object> fact = facts.getFact();
				if (c.equals(Future.class)) {
					if (facts.getReturnType() != null)
						c = facts.getReturnType();
					DroolsFutureTask<T> dft = new DroolsFutureTask<T>(this,
							proxy, (facts.getReturnType() != null) ? c : null,
							method, session, args, fact);
					executors.execute(dft);
					return dft;
				}
				if (facts.getReturnType() != null)
					c = facts.getReturnType();
				return invokeInternal(proxy, method, c, session, args, fact);
			}
		} else if (method.getDeclaringClass().equals(DroolsInterface.class)) {
			return method.invoke(inter, args);
		}
		String msg = "Unable to handle the received parameter";
		logger.error(msg);
		throw new IllegalArgumentException(msg);
	}

	public Object invokeInternal(Object proxy, Method method,
			Class<?> returnType, DroolsSessionType session, Object[] args,
			List<Object> fact) throws Throwable {
		Iterator<?> result = null;
		if (DroolsSessionType.NONE.equals(session)) {
			result = inter.runStatelessSession(fact);
		} else if (DroolsSessionType.NEW.equals(session)) {
			if (inter.getCurrentSession() != null) {
				inter.getCurrentSession().halt();
				inter.setCurrentSession(null);
			}
			result = inter.runStatefulSession(fact, false);
		} else if (DroolsSessionType.END.equals(session)) {
			result = inter.runStatefulSession(fact, true);
			inter.getCurrentSession().halt();
			inter.setCurrentSession(null);
		} else if (DroolsSessionType.JOIN.equals(session)) {
			if (inter.getCurrentSession() == null)
				result = inter.runStatelessSession(fact);
			else
				result = inter.runStatefulSession(fact, true);
		} else if (DroolsSessionType.REQUIRE.equals(session)) {
			result = inter.runStatefulSession(fact, true);
		}
		return inter.getResult(returnType, result);
	}

	private Class<T> clazz;

	private DroolsInterfaceImplementation<T> inter;

	public DroolsProxy(Class<T> clazz, DroolsInterfaceImplementation<T> inter) {
		this.clazz = clazz;
		this.inter = inter;
	}

	private static final ExecutorService executors = Executors
			.newCachedThreadPool();

}
